/**
 * Copyright (c) 2020 Bubble, Inc.  All rights reserved.
 * For personal (non-commercial) use, see license: https://getbubblenow.com/bubble-license/
 */
package bubble.service.backup;

import bubble.cloud.storage.StorageServiceDriver;
import bubble.dao.cloud.CloudServiceDAO;
import bubble.model.cloud.BubbleBackup;
import bubble.model.cloud.CloudCredentials;
import bubble.model.cloud.CloudService;
import bubble.model.cloud.NetworkKeys;
import bubble.server.BubbleConfiguration;
import lombok.Cleanup;
import lombok.Getter;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.exec.CommandLine;
import org.apache.commons.io.IOUtils;
import org.cobbzilla.util.collection.NameAndValue;
import org.cobbzilla.util.system.Bytes;
import org.cobbzilla.util.system.Command;
import org.cobbzilla.wizard.cache.redis.RedisService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;

import static bubble.ApiConstants.HOME_DIR;
import static bubble.model.cloud.NetworkKeys.PARAM_STORAGE;
import static bubble.model.cloud.NetworkKeys.PARAM_STORAGE_CREDENTIALS;
import static java.io.File.createTempFile;
import static java.util.concurrent.TimeUnit.HOURS;
import static java.util.concurrent.TimeUnit.MINUTES;
import static org.cobbzilla.util.daemon.ZillaRuntime.die;
import static org.cobbzilla.util.io.FileUtil.*;
import static org.cobbzilla.util.json.JsonUtil.json;
import static org.cobbzilla.util.security.CryptStream.BUFFER_SIZE;
import static org.cobbzilla.util.security.ShaUtil.sha256_hex;
import static org.cobbzilla.util.system.CommandShell.exec;
import static org.cobbzilla.util.system.CommandShell.okResult;
import static org.cobbzilla.wizard.cache.redis.RedisService.EX;

@Service @Slf4j
public class RestoreService {

    public static final File RESTORE_DIR = new File(HOME_DIR, "restore");

    // this is how long the RestoreService will allow a restore, after the API starts
    public static final long RESTORE_WINDOW = HOURS.toMillis(24);
    public static final long RESTORE_WINDOW_SECONDS = RESTORE_WINDOW/1000;

    // this is how long bubble_restore_monitor.sh will allow a restore after it starts
    // we add some time because, in the ansible setup, the script starts (in role bubble) before the
    // API is started (in role finalizer)
    public static final long RESTORE_MONITOR_SCRIPT_TIMEOUT_SECONDS = RESTORE_WINDOW_SECONDS + MINUTES.toSeconds(5);

    private static final long RESTORE_LOCK_TIMEOUT = MINUTES.toMillis(31);
    private static final long RESTORE_DEADLOCK_TIMEOUT = MINUTES.toMillis(30);
    private static final File RESTORE_MARKER_FILE = new File(HOME_DIR, ".restore");

    private static final int BACKUP_ARCHIVE_MANAGEMENT_BUFFER_SIZE = (int) (8 * Bytes.KB);

    @Autowired private CloudServiceDAO cloudDAO;

    @Autowired private RedisService redis;
    @Getter(lazy=true) private final RedisService restoreKeys = redis.prefixNamespace(getClass().getSimpleName());

    @Autowired private BubbleConfiguration configuration;

    public void registerRestore(String restoreKey, NetworkKeys keys) {
        getRestoreKeys().set(restoreKey, json(keys), EX, RESTORE_WINDOW_SECONDS);
    }

    public boolean isValidRestoreKey(String restoreKey) { return getRestoreKeys().exists(restoreKey); }

    public boolean isSelfRestoreStarted() {
        // only one restore may be started, hence
        return getRestoreKeys().keys("*").size() > 0 || RESTORE_MARKER_FILE.exists();
    }

    public boolean restore(String restoreKey, BubbleBackup backup) {
        final String thisNodeUuid = configuration.getThisNode().getUuid();
        final String thisNetworkUuid = configuration.getThisNode().getNetwork();
        String lock = null;
        try {
            lock = getRestoreKeys().lock(thisNetworkUuid, RESTORE_LOCK_TIMEOUT, RESTORE_DEADLOCK_TIMEOUT);

            final String keyJson = getRestoreKeys().get(restoreKey);
            if (keyJson == null) {
                log.error("restore: restoreKey not found: " + restoreKey);
                return false;
            }
            final NetworkKeys networkKeys = json(keyJson, NetworkKeys.class);
            final String storageJson = NameAndValue.find(networkKeys.getKeys(), PARAM_STORAGE);
            final String credentialsJson = NameAndValue.find(networkKeys.getKeys(), PARAM_STORAGE_CREDENTIALS);
            if (storageJson == null || credentialsJson == null) {
                log.error("restore: storage/credentials not found in NetworkKeys");
                return false;
            }

            final String[] existingFiles = RESTORE_DIR.list();
            if (existingFiles != null && existingFiles.length > 0) {
                log.error("restore: files already exist in " + abs(RESTORE_DIR) + ", cannot restore");
                return false;
            }

            final CloudService storageService = json(storageJson, CloudService.class)
                    .setCredentials(json(credentialsJson, CloudCredentials.class));
            try {
                fetchBackupFiles(thisNodeUuid, storageService, backup, RESTORE_DIR);
                log.info("restore: notifying system to restore from backup at: "+abs(RESTORE_DIR));
                touch(RESTORE_MARKER_FILE);
                return true;
            } catch (IOException e) {
                log.error("restore: error downloading backup ", e);
                return false;
            }
        } finally {
            if (lock != null) {
                getRestoreKeys().unlock(thisNetworkUuid, lock);
            }
        }
    }

    /**
     * Build temporal ZIP with (specified) backup files for specified network/node.
     * @return Created tar file with backup.
     */
    @NonNull public File buildRestorePackage(@NonNull final String nodeUuid, @NonNull final BubbleBackup backup,
                                             @NonNull final String passphrase) {
        final var storageService = cloudDAO.findByUuid(configuration.getThisNetwork().getStorage());
        if (storageService == null) return die("buildRestorePackage: cannot find network storage for this network");

        try {
            final var backupDir = fetchBackupFiles(nodeUuid, storageService, backup, createTempDir("backup-"));
            final var backupPackage = createTempFile("backup-", ".tgz.enc");
            final var cmd = new CommandLine(
                    "cd " + backupDir.getAbsolutePath()
                    + " && tar -cz *"
                    + " | openssl enc -aes-256-cbc -pbkdf2 -iter 10000 -pass pass:" + sha256_hex(passphrase)
                    + " > " + backupPackage.getAbsolutePath()
                    + " && rm -rf *"); // removing the separate backup files
            okResult(exec(new Command(cmd)));
            return backupPackage;
        } catch (Exception e) {
            return die("buildRestorePackage: error downloading backup files ", e);
        }
    }

    @NonNull private File fetchBackupFiles(@NonNull final String thisNodeUuid,
                                           @NonNull final CloudService storageService,
                                           @NonNull final BubbleBackup backup,
                                           @NonNull final File destDir)
            throws IOException {
        final var path = StorageServiceDriver.getPath(backup.getPath());
        log.info("fetchBackupFiles: downloading backup from path=" + path);
        final var storageDriver = storageService.getStorageDriver(configuration);
        final var destDirAbs = abs(destDir);

        var listing = storageDriver.list(thisNodeUuid, path);
        while (true) {
            Arrays.stream(listing.getKeys())
                  .parallel()
                  .forEach(k -> {
                      final var logMsgPrefix = "fetchBackupFiles [" + k + "]: ";
                      log.info(logMsgPrefix + "downloading file");
                      final var file = new File(destDirAbs, k);
                      mkdirOrDie(file.getParentFile());
                      try {
                          @Cleanup final var in = storageDriver.read(thisNodeUuid, k);
                          try (var out = new BufferedOutputStream(new FileOutputStream(file), BUFFER_SIZE)) {
                              IOUtils.copyLarge(in, out);
                          }
                          log.info(logMsgPrefix + "successfully downloaded file");
                      } catch (Exception e) {
                          die(logMsgPrefix + "error downloading file", e);
                      }
                  });

            if (!listing.isTruncated()) break;
            listing = storageDriver.listNext(thisNodeUuid, listing.getListingId());
        }
        log.info("fetchBackupFiles: full download successful");
        return destDir;
    }
}
